import xbmc
import xbmcaddon
import xbmcgui
import os
import platform
import socket
import subprocess
import time


__addon__        = xbmcaddon.Addon()
__addonversion__ = __addon__.getAddonInfo('version')
__addonid__      = __addon__.getAddonInfo('id')
__addonname__    = __addon__.getAddonInfo('name')
__localize__     = __addon__.getLocalizedString
__working_dir__  = __addon__.getAddonInfo('path').decode('utf-8')
__architecture__ = platform.machine()


def MooseFS_Log(txt):
    message = '%s: %s' % (__addonname__, txt.encode('ascii', 'ignore'))
    xbmc.log(msg=message, level=xbmc.LOGNOTICE)


class MooseFS_Main:

    _debug = False
    _mount_bin = ""
    _umount_bin = "umount"
    _default_settings = {
            'mount_point': '',
            'moosefs_host': '',
            'moosefs_port': 9421,
            'moosefs_path': '',
            'moosefs_password': '',
            'o_read_write': True,
            'o_suid': False,
            'o_dev': False,
            'o_exec': False,
            'mount_timeout': 3,
            'quiet': False
        }
    _cfg_items = dict()
    _old_items = dict()
    _is_mounted = False
    _remount_settings = [ 
            'mount_point',
            'moosefs_host',
            'moosefs_port',
            'moosefs_path',
            'moosefs_password',
            'o_read_write',
            'o_suid',
            'o_dev',
            'o_exec'
        ]

    def __init__(self):
        os.environ['PYTHONUNBUFFERED'] = "1"
        if __architecture__.startswith("arm"):
            self._mount_bin = os.path.join(__working_dir__, 'resources/mfsmount-armhf')
        else:
            self._mount_bin = os.path.join(__working_dir__, 'resources/mfsmount-amd64')
        self._old_items = dict(self._default_settings)
        self._cfg_items = dict(self._default_settings)
        self._settings_resetter(invalid_items=self._settings_get())
        if os.getuid() != 0:
            self._notify_user(lvl=2, secs=8,
                msg='Not running as a privileged user (id={0})'.format(os.getuid()))
            xbmc.sleep(7000)                       
        if not os.access(self._mount_bin, os.X_OK):
            os.chmod(self._mount_bin, 0o755)                                              
        if self._mount_viable():
            self._mount()
        self.Monitor = MooseFS_Monitor(update_settings = self._settings_change)
        self._daemon()

    def _daemon(self):
        while not self.Monitor.abortRequested():
            if self.Monitor.waitForAbort(600):
                break
            self._mount_status()
        self.Monitor.update_settings = None
        self._umount()

    def _notify_user(self, *args, **kwargs):
        if not 'msg' in kwargs:
            return False
        msg = kwargs['msg']
        level = xbmcgui.NOTIFICATION_INFO
        if 'lvl' in kwargs:
            if kwargs['lvl'] == 2:
                level = xbmcgui.NOTIFICATION_WARNING
            if kwargs['lvl'] == 3:
                level = xbmcgui.NOTIFICATION_ERROR
        timeout = 5000
        if 'secs' in kwargs:
            timeout = int(kwargs['secs']) * 1000
        if self._cfg_items['quiet']:
            MooseFS_Log('User not notified: {0}'.format(msg))
        else:
            dialog = xbmcgui.Dialog()
            dialog.notification('MooseFS Client',
                msg, level, timeout)
            MooseFS_Log('User notified: {0}'.format(msg))

    def _validate_host(self, *args, **kwargs):
        if not 'value' in kwargs:
            return False
        hostname = kwargs['value']
        try:
            if socket.gethostbyname(hostname) == hostname:
                return True
            elif socket.gethostbyname(hostname) != hostname:
                return True
        except socket.gaierror:
            return False
        return False

    def _validate_port(self, *args, **kwargs):
        if not 'value' in kwargs:
            return False
        port = kwargs['value']
        if port > 0 and port < 65536:
            return True
        return False

    def _validate_path(self, *args, **kwargs):
        if not 'value' in kwargs:
            return False
        port = kwargs['value']
        if kwargs['value'] == os.path.abspath(kwargs['value']):
            return True
        return False

    def _validate_settings(self):
        invalid_items = [ ]
        if not xbmc.validatePath(self._cfg_items['mount_point']):
            invalid_items.append('mount_point')
        if not self._validate_host(value=self._cfg_items['moosefs_host']):
            invalid_items.append('moosefs_host')
        if not self._validate_port(value=self._cfg_items['moosefs_port']):
            invalid_items.append('moosefs_port')
        if not self._validate_path(value=self._cfg_items['moosefs_path']):
            invalid_items.append('moosefs_path')
        return invalid_items

    def _settings_resetter(self, *args, **kwargs):
        if not 'invalid_items' in kwargs:
            return
        invalid_items = kwargs['invalid_items']
        if not len(invalid_items):
            return
        interactive = True if 'interactive' in kwargs else False
        if self._debug:
            if interactive and not self._cfg_items['quiet']:
                self._notify_user(lvl=2,
                    msg='Invalid settings are assumed as defaults')
        for config_item in invalid_items:
            self._cfg_items[config_item] = self._default_settings[config_item]
            # __addon__.setSetting(config_item, self._cfg_items[config_item])
        return

    def _should_remount(self):
        if self._debug:
            MooseFS_Log('cfg: {0}'.format(self._cfg_items))
            MooseFS_Log('old: {0}'.format(self._old_items))
        for item in self._remount_settings:
            if not item in self._cfg_items:
                return True
            if not item in self._old_items:
                return True
            if self._cfg_items[item] != self._old_items[item]:
                return True
        return False

    def _settings_get(self):
        self._cfg_items['mount_point'] = xbmc.translatePath(__addon__.getSetting("mount_point"))
        self._cfg_items['moosefs_host'] = __addon__.getSettingString("moosefs_host")
        self._cfg_items['moosefs_port'] = __addon__.getSettingInt("moosefs_port")
        self._cfg_items['moosefs_path'] = __addon__.getSettingString("moosefs_path")
        self._cfg_items['moosefs_password'] = __addon__.getSettingString("moosefs_password")
        self._cfg_items['o_read_write'] = __addon__.getSettingBool("o_read_write")
        self._cfg_items['o_suid'] = __addon__.getSettingBool("o_suid")
        self._cfg_items['o_dev'] = __addon__.getSettingBool("o_dev")
        self._cfg_items['o_exec'] = __addon__.getSettingBool("o_exec")
        self._cfg_items['mount_timeout'] = __addon__.getSettingInt("mount_timeout")
        self._cfg_items['quiet'] = __addon__.getSettingBool("quiet")
        return self._validate_settings()

    def _settings_change(self):
        if self._debug:
            MooseFS_Log('settings changed ..')
        self._old_items = dict(self._cfg_items)
        self._settings_resetter(invalid_items=self._settings_get(), interactive=True)
        if not self._should_remount():
            return
        if self._is_mounted:
            if self._cfg_items['mount_point'] != self._old_items['mount_point']:
                self._umount(mount_point=self._old_items['mount_point'])
            self._umount(mount_point=self._cfg_items['mount_point'])
        if not self._mount_viable():
            self._notify_user(lvl=3,
                msg='Not mounted: provide mount point and host')
            return
        self._mount()

    def _mount_status(self):
        if self._cfg_items['mount_point']:
	    if os.path.ismount(self._cfg_items['mount_point']):
                self._is_mounted = True
                return True
        self._is_mounted = False
        return False

    def _mount_viable(self):
        if self._cfg_items['mount_point'] and self._cfg_items['moosefs_host']:
            return True
        return False

    def _mount(self):
        if self._mount_status():
            self._umount()
	    if os.path.ismount(self._cfg_items['mount_point']):
                self._is_mounted = True
                dialog = xbmcgui.Dialog()
                self._notify_user(lvl=3,
                    msg='Mount point is busy (mounted?)')
                return False
        if not self._mount_viable():
            return False

        params = [ ]
        params.append(self._mount_bin)
        params.append(self._cfg_items['mount_point'])
        params.append('-H')
        params.append(self._cfg_items['moosefs_host'])
        if self._cfg_items['moosefs_port']:
            params.append('-P')
            params.append(str(self._cfg_items['moosefs_port']))
        if self._cfg_items['moosefs_path']:
            params.append('-S')
            params.append(self._cfg_items['moosefs_path'])
        if self._cfg_items['moosefs_password']:
            params.append('-o')
            params.append('mfspassword=' + self._cfg_items['moosefs_password'])
        params.append('-o')
        if self._cfg_items['o_read_write']:
            params.append('rw')
        else:
            params.append('ro')
        params.append('-o')
        if self._cfg_items['o_suid']:
            params.append('suid')
        else:
            params.append('nosuid')
        params.append('-o')
        if self._cfg_items['o_dev']:
            params.append('dev')
        else:
            params.append('nodev')
        params.append('-o')
        if self._cfg_items['o_exec']:
            params.append('exec')
        else:
            params.append('noexec')

        if self._debug:
            MooseFS_Log('Mounting MooseFS on %s' % self._cfg_items['mount_point'])
        self._exec(parms=params, to=self._cfg_items['mount_timeout'])

        if os.path.ismount(self._cfg_items['mount_point']):
            self._notify_user(lvl=1,
                msg='Successfully mounted volume')
            self._is_mounted = True
            return True

        self._notify_user(lvl=2, msg='Failed to mount volume')
        self._is_mounted = False
        return False

    def _umount(self, *args, **kwargs):
        umount_point = self._cfg_items['mount_point']
        if 'mount_point' in kwargs:
            umount_point = kwargs['mount_point']

        if len(umount_point) == 0:
            self._is_mounted = False
            return False
	if not os.path.ismount(umount_point):
            self._is_mounted = False
            return True

        params = [ ]
        params.append('umount')
        params.append(umount_point)

        if self._debug:
            MooseFS_Log('Umounting MooseFS on {0}'.format(umount_point))
        self._exec(parms=params, to=self._cfg_items['mount_timeout'])

	if not os.path.ismount(umount_point):
            self._notify_user(lvl=1,
                msg='Successfully unmounted volume')
            self._is_mounted = False
            return True

        self._notify_user(lvl=2, msg='Failed to unmount volume')
        self._is_mounted = True
        return False

    def _exec(self, *args, **kwargs):
        if not 'parms' in kwargs or not 'to' in kwargs:
            return False
        parameters = kwargs['parms']
        timeout = kwargs['to'] + 1
        if self._debug:
            MooseFS_Log('Command to execute {0}'.format(kwargs['parms']))

        return_value = None
        clock_start = time.time()
        ph = subprocess.Popen(parameters,
            stdout = subprocess.PIPE,
            stderr = subprocess.STDOUT)
        sub_out, sub_err = ph.communicate()
        while True:
            return_value = ph.poll()
            if self._debug:
                MooseFS_Log('Command {0}:{1}'.format(time.time(),return_value))
            if return_value != None:
                break
            if time.time() - clock_start >= timeout:
                try:
                    ph.kill()
                except OSError:
                    pass
                self._notify_user(lvl=3, msg='Command execution timed-out')
                break
            xbmc.sleep(100)
        if self._debug:
            if self._debug:
                MooseFS_Log('Exec output: {0}'.format(sub_out))
        return return_value


class MooseFS_Monitor(xbmc.Monitor):
    def __init__(self, *args, **kwargs):
        xbmc.Monitor.__init__(self)
        self.update_settings = kwargs['update_settings']

    def onSettingsChanged(self):
        self.update_settings()


if __name__ == '__main__':
    monitor = xbmc.Monitor()
    MooseFS_Log('Version {0} started'.format(__addonversion__))
    MooseFS_Main()
    del MooseFS_Monitor
    del MooseFS_Main
    MooseFS_Log('Version {0} stopped'.format(__addonversion__))

